/*  Termite filter: Transmit user-defined text on a function key
 *
 *  Copyright (c) ITB CompuPhase, 2010-2018
 *
 *  This software is provided "as-is", without any express or implied warranty.
 *  In no event will the authors be held liable for any damages arising from
 *  the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, subject to the following restrictions:
 *
 *  1.  The origin of this software must not be misrepresented; you must not
 *      claim that you wrote the original software. If you use this software in
 *      a product, an acknowledgment in the product documentation would be
 *      appreciated but is not required.
 *  2.  Altered source versions must be plainly marked as such, and must not be
 *      misrepresented as being the original software.
 *  3.  This notice may not be removed or altered from any source distribution.
 */

#include <assert.h>
#include <windows.h>
#include <mmsystem.h>
#include "functionkeys.h"


#define FLTEXPORT __declspec(dllexport) __stdcall

#if !defined sizearray
  #define sizearray(a)  (sizeof(a) / sizeof((a)[0]))
#endif

    /* a few strings that are used in various locations */
const char strCaption[] = "FunctionKeys filter";
const char strSection[] = "FunctionKeys";
const char strKeyItem[] = "key%d";
const char strKeyItemRepeat[] = "key%d-repeat";


    /* forward declarations of a few functions */
BOOL CALLBACK FilterConfigDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK FilterInfoDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static void reload(void);
static BOOL strparse(char *output, int size, const char *input);
static int gcd(int a, int b);
static void CALLBACK TimerCallbackWindow(UINT idTimer,UINT msg,DWORD_PTR dwUser,
                                         DWORD_PTR dw1,DWORD_PTR dw2);

    /* constants */
#define MAX_FKEYS       12  /* support up to 12 function keys right-now */
#define MAX_TEXTLENGTH  256 /* max. length of macro text */
#define UM_POSTHOTKEY   (WM_APP + 3)  /* also defined in filter.h */

    /* this variable is declared in dllmain.c */
extern HINSTANCE hinstDLL;

    /* a few global variables for this particular filter */
typedef struct tagKEYMACRO {
  char *Text;
  DWORD Repeat;
  DWORD RepeatTimeout;
} KEYMACRO;
static KEYMACRO KeyDefs[MAX_FKEYS];
static char ProfileFile[_MAX_PATH];
static HWND hwndTermite;
static int TimerInterval;
static MMRESULT TimerID = 0;


/* flt_Load() is called when Termite loads the filter.
 *
 * hwnd         The handle to the main window of Termite.
 * ProfileName  The full path to the INI file that Termite uses. The filter
 *              can use this name to store its configuration.
 * Build        The build number of Termite, for distinguishing versions of
 *              the Termite application.
 *
 * This function must return TRUE if it can load successfully. If the function
 * returns FALSE, Termite will unload it.
 */
BOOL FLTEXPORT flt_Load(HWND hwnd, LPCSTR ProfileName, int Build)
{
  int idx;

  (void)Build;        /* this parameter is not used */
  hwndTermite = hwnd;
  strcpy(ProfileFile, ProfileName); /* save name of the INI file */

  /* clear a few globals before "reload" */
  TimerID = 0;
  for (idx = 0; idx < MAX_FKEYS; idx++)
    KeyDefs[idx].Text = NULL;
  reload();

  return TRUE;
}


/* flt_Unload() is called when Termite removes the filter.
 *
 * After calling this notification function, Termite will physically unload
 * the filter DLL.
 *
 * For filters that allocate dynamic memory or other resources, this is a good
 * moment to free these resources.
 */
void FLTEXPORT flt_Unload(void)
{
  int idx;

  if (TimerID != 0) {
    timeKillEvent(TimerID);
    timeEndPeriod(1);
  }

  /* free macros */
  for (idx = 0; idx < MAX_FKEYS; idx++) {
    if (KeyDefs[idx].Text != NULL)
      free(KeyDefs[idx].Text);
    KeyDefs[idx].RepeatTimeout = 0;
  }
}


/* flt_HotKey() is called on a function key press.
 *
 * vKey         The virtual key of the function key.
 * Text         A pointer to a buffer where to store the data to transmit.
 *              On entry, this string is empty, unless another filter has
 *              generated data for this function key as well.
 * Size         This parameter must hold the new size of the data block (on
 *              output).
 *
 * If flt_HotKey() does not produce output, it can return NULL or return
 * the original (usually empty) string.
 *
 * The output string is not necessarily zero-terminated; the filter must adjust
 * the length to the number of bytes it returns (parameter "Size").
 */
LPCSTR FLTEXPORT flt_HotKey(int vKey, LPCSTR Text, LPINT Size)
{
  int idx;

  /* find the entry */
  idx = vKey - VK_F1;
  if (idx < 0 || idx >= MAX_FKEYS)
    return NULL;        /* function key is out of range */

  /* check that we handle this key */
  if (KeyDefs[idx].Text == NULL || strlen(KeyDefs[idx].Text) == 0)
    return NULL;

  /* verify whether we are the first filter to handle this key */
  if (strlen(Text) > 0) {
    MessageBox(hwndTermite, "Duplicate definition for this function key", NULL, MB_OK);
    return NULL;
  } /* if */

  /* as soon as the key is pressed, flag it for auto-repeat (or turn it off) */
  if ((GetKeyState(VK_SHIFT) & 0x80000000) == 0)
    KeyDefs[idx].RepeatTimeout = KeyDefs[idx].Repeat;
  else
    KeyDefs[idx].RepeatTimeout = 0;

  *Size = strlen(KeyDefs[idx].Text);
  return KeyDefs[idx].Text;
}


/* flt_Config() is called when the user chooses to configure the filter.
 *
 * This function must return TRUE if the configuration was successfully changed,
 * and FALSE if the user cancelled the operation or if there was an error.
 */
BOOL FLTEXPORT flt_Config(void)
{
  /* pop up the dialog for configuration */
  int result = DialogBox(hinstDLL, "FilterConfig", GetFocus(), (DLGPROC)FilterConfigDlg);
  if (result == IDOK)
    reload();   /* the dialog has saved the settings in the INI file, reload */

  return (result == IDOK);
}


#pragma argsused
BOOL CALLBACK FilterConfigDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  char str[MAX_TEXTLENGTH], key[32];
  int idx, value;

  switch (msg) {
  case WM_INITDIALOG:
    /* load current settings from the INI file, without translation */
    for (idx = 0; idx < MAX_FKEYS; idx++) {
      wsprintf(key, strKeyItem, idx + 1);
      GetPrivateProfileString(strSection, key, "", str, sizearray(str), ProfileFile);
      SetDlgItemText(hwnd, EDIT_F1 + idx, str);
      SendDlgItemMessage(hwnd, EDIT_F1 + idx, EM_LIMITTEXT, MAX_TEXTLENGTH - 1, 0);
      wsprintf(key, strKeyItemRepeat, idx + 1);
      value = GetPrivateProfileInt(strSection, key, 0, ProfileFile);
      if (value > 0) {
        CheckDlgButton(hwnd, CHK_REPEAT_F1 + idx, BST_CHECKED);
        SetDlgItemInt(hwnd, EDIT_REPEAT_F1 + idx, value, FALSE);
      } else {
        EnableWindow(GetDlgItem(hwnd, EDIT_REPEAT_F1 + idx), FALSE);
      }
    } /* for */
    return TRUE;

  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDCANCEL:
      EndDialog(hwnd, LOWORD(wParam));
      return TRUE;
    case IDOK:
      /* save the fields into the INI file */
      for (idx = 0; idx < MAX_FKEYS; idx++) {
        GetDlgItemText(hwnd, EDIT_F1 + idx, str, sizearray(str));
        wsprintf(key, strKeyItem, idx + 1);
        if (strlen(str) > 0)
          WritePrivateProfileString(strSection, key, str, ProfileFile);
        else
          WritePrivateProfileString(strSection, key, NULL, ProfileFile);
        if (IsDlgButtonChecked(hwnd, CHK_REPEAT_F1 + idx))
          value = GetDlgItemInt(hwnd, EDIT_REPEAT_F1 + idx, NULL, FALSE);
        else
          value = 0;
        wsprintf(key, strKeyItemRepeat, idx + 1);
        if (value > 0) {
          wsprintf(str, "%d", value);
          WritePrivateProfileString(strSection, key, str, ProfileFile);
        } else {
          WritePrivateProfileString(strSection, key, NULL, ProfileFile);
        }
      }
      EndDialog(hwnd, LOWORD(wParam));
      return TRUE;
    case IDHELP:
      DialogBox(hinstDLL, "FilterInfo", hwnd, (DLGPROC)FilterInfoDlg);
      return TRUE;
    case CHK_REPEAT_F1:
    case CHK_REPEAT_F2:
    case CHK_REPEAT_F3:
    case CHK_REPEAT_F4:
    case CHK_REPEAT_F5:
    case CHK_REPEAT_F6:
    case CHK_REPEAT_F7:
    case CHK_REPEAT_F8:
    case CHK_REPEAT_F9:
    case CHK_REPEAT_F10:
    case CHK_REPEAT_F11:
    case CHK_REPEAT_F12:
      idx = LOWORD(wParam) - CHK_REPEAT_F1;
      EnableWindow(GetDlgItem(hwnd, EDIT_REPEAT_F1 + idx), IsDlgButtonChecked(hwnd, CHK_REPEAT_F1 + idx));
      return TRUE;
    } /* switch */
    break;
  } /* switch */

  return FALSE;
}

#pragma argsused
BOOL CALLBACK FilterInfoDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
static
#include "functionkeys-help.c"

  switch (msg) {
  case WM_INITDIALOG:
    SetDlgItemText(hwnd, INFOTEXT, helpstring);
    return TRUE;
  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDCANCEL:
    case IDOK:
      EndDialog(hwnd, LOWORD(wParam));
      return TRUE;
    } /* switch */
  } /* switch */

  return FALSE;
}


static void reload(void)
{
  char cooked[MAX_TEXTLENGTH], raw[MAX_TEXTLENGTH];
  char key[32];
  int idx;

  /* clear / free old data */
  if (TimerID != 0) {
    timeKillEvent(TimerID);
    timeEndPeriod(1);
    TimerID = 0;
  }
  for (idx = 0; idx < MAX_FKEYS; idx++) {
    if (KeyDefs[idx].Text != NULL)
      free(KeyDefs[idx].Text);
    KeyDefs[idx].RepeatTimeout = 0;
  }

  /* read settings (from the same INI file as Termite uses) */
  for (idx = 0; idx < MAX_FKEYS; idx++) {
    wsprintf(key, strKeyItem, idx + 1);
    GetPrivateProfileString(strSection, key, "", raw, sizearray(raw), ProfileFile);
    strparse(cooked, sizearray(cooked), raw);
    KeyDefs[idx].Text = strdup(cooked);
    wsprintf(key, strKeyItemRepeat, idx + 1);
    KeyDefs[idx].Repeat = GetPrivateProfileInt(strSection, key, 0, ProfileFile);
  }

  /* see whether a timer must be set up */
  TimerInterval = 0;
  for (idx = 0; idx < MAX_FKEYS; idx++) {
    if (KeyDefs[idx].Repeat > 0) {
      /* calculate the smallest common interval */
      if (TimerInterval == 0)
        TimerInterval = KeyDefs[idx].Repeat;
      else
        TimerInterval = gcd(TimerInterval, KeyDefs[idx].Repeat);
    }
  }
  if (TimerInterval > 0) {
    timeBeginPeriod(1);
    TimerID = timeSetEvent(TimerInterval, 0, TimerCallbackWindow, 0, TIME_PERIODIC);
  }
}

static int hexchar(char ch)
{
  if ('0' <= ch && ch <='9')
    return ch - '0';
  if ('a' <= ch && ch <='f')
    return ch - 'a' + 10;
  if ('A' <= ch && ch <='F')
    return ch - 'A' + 10;
  return -1;
}

/* strparse() copies a string from input to output and translates escaped
 * characters to the original characters. The return value is whether all escape
 * sequences are valid, and if the complete string was copied. If there are one
 * or more invalid escape sequences, the returned result is FALSE.
 */
static BOOL strparse(char *output, int size, const char *input)
{
  BOOL result = TRUE;
  int count = 0;

  assert(output != NULL);
  assert(size > 0);
  assert(input != NULL);
  while (*input != '\0' && count < size) {
    if (*input == '\\') {
      const char *skip = input + 1;
      char byte = 0;
      switch (*skip) {
      case 'n':
        byte = '\n';
        break;
      case 'r':
        byte = '\r';
        break;
      case 't':
        byte = '\t';
        skip++;
        break;
      case 'x':
        if (hexchar(skip[1]) >= 0 && hexchar(skip[2]) >= 0) {
          byte = (hexchar(skip[1]) << 4) | hexchar(skip[2]);
          skip += 2;
        }
        break;
      case '\\':
        byte = '\\';
        break;
      }
      if (byte != 0) {
        *output = byte;
        input = skip;
      } else {
        *output = *input;
        result = FALSE;
      }
    } else {
      *output = *input;
    }
    input++;
    output++;
    count++;
  }

  /* zero-terminate */
  if (count < size)
    *output = '\0';
  else
    result = FALSE;

  return result;
}

/* The greatest common divisor of two values, using Euclides' algorithm */
static int gcd(int a, int b)
{
  while (a != b) {
    if (a > b)
      a = a - b;
    else
      b = b - a;
  }
  return a;
}

#pragma argsused
static void CALLBACK TimerCallbackWindow(UINT idTimer,UINT msg,DWORD_PTR dwUser,
                                         DWORD_PTR dw1,DWORD_PTR dw2)
{
  int idx;

  for (idx = 0; idx < MAX_FKEYS; idx++) {
    if (KeyDefs[idx].RepeatTimeout > 0) {
      KeyDefs[idx].RepeatTimeout -= TimerInterval;
      if (KeyDefs[idx].RepeatTimeout == 0) {
        if (KeyDefs[idx].Text != NULL && strlen(KeyDefs[idx].Text) != 0)
          PostMessage(hwndTermite, UM_POSTHOTKEY, strlen(KeyDefs[idx].Text), (long)KeyDefs[idx].Text);
        KeyDefs[idx].RepeatTimeout = KeyDefs[idx].Repeat; /* reload repeat interval */
      }
    }
  }
}
